#include <stdlib.h> /* calloc */
#include <stdio.h>  /* printf */
#include <assert.h>
#include <sys/time.h>
#include <string.h> /* memcpy, memset */
#include <cuda.h>
#include <cutil_inline.h> /* CUDA_SAFE_CALL */
#include "../../common/common.h"

#define KERNEL_FLOP 4
#define KERNEL_MOP 5

/* Device code */
__global__ void heat_red(float *mtx)
{
	int i = blockIdx.y*TILE + threadIdx.y;
	int j = blockIdx.x*TILE + threadIdx.x;
	i = 2*i + j%2;
	if (0<i && i<N-1 && 0<j && j<N-1) {
			mtx[INDEX(i,j)] = 0.25f* (
					mtx[INDEX(i-1,j)]+
				mtx[INDEX(i,j-1)] + mtx[INDEX(i,j+1)] +
					mtx[INDEX(i+1,j)]
						);
	}
}

__global__ void heat_black(float *mtx)
{
	int i = blockIdx.y*TILE + threadIdx.y;
	int j = blockIdx.x*TILE + threadIdx.x;
	i = 2*i + (j+1)%2;
	if (0<i && i<N-1 && 0<j && j<N-1) {
			mtx[INDEX(i,j)] = 0.25f* (
					mtx[INDEX(i-1,j)]+
				mtx[INDEX(i,j-1)] + mtx[INDEX(i,j+1)] +
					mtx[INDEX(i+1,j)]
						);
	}
}


/* Host code */
int main(int argc, char** argv)
{
	unsigned int i = 0;
	dim3 threadsPerBlock(TILE,TILE,1);
	dim3 blocksPerGrid(DIV_CEIL(N,TILE),DIV_CEIL(N/2,TILE),1);

	/* At least a border square surrounding a single cell */
	assert(2<=N);
	/* Mtx fit into VRAM */
	assert(N*N*sizeof(float) <= (1<<30));
	/* Even number of rows for checkboard splitting */
	assert(N%2==0);

	/* Enough threads */
	assert(N<=blocksPerGrid.x*threadsPerBlock.x);
	assert(N/2<=blocksPerGrid.y*threadsPerBlock.y);
	/* Hardware limits */
	assert(threadsPerBlock.x*threadsPerBlock.y<=512);
	assert(blocksPerGrid.x<=65535 && blocksPerGrid.y<=65535);

	/* Set device to use */
	CUDA_SAFE_CALL(cudaSetDevice(DEVICE));

	/* Allocate vectors in device memory */
	size_t size = N*N * sizeof(float);
	assert(0<size);

	float *h_mtx = NULL;
	float *d_mtx = NULL;

	/* Alloc and set host and device memory */
	h_mtx = (float *)calloc(1, size);
	assert(h_mtx);
	CUDA_SAFE_CALL(cudaMalloc((void **)&d_mtx, size));
	assert(d_mtx);

	for (i=0; i<N; i++) {
		float value = HOT - (4.0/(N*N) * HOT)*(N/2.0-i)*(N/2.0-i);
		h_mtx[INDEX(0,i)] = h_mtx[INDEX(N-1,i)] = value;
		h_mtx[INDEX(i,0)] = h_mtx[INDEX(i,N-1)] = value;
	}

	/* There is no temp one pixel off the center */
	assert(0.0==h_mtx[OFFBYONE_CORNER]);
	/* Before */
	writePPMbinaryImage("image_before.ppm", h_mtx);

	/* Copy vector from host memory to device memory */
	CUDA_SAFE_CALL(cudaMemcpy(d_mtx, h_mtx, size, cudaMemcpyHostToDevice));

	/* Start timer */
	double secs = 0.0;
	struct timeval start = {0L,0L}, end = {0L,0L}, elapsed = {0L,0L};
	gettimeofday(&start,NULL);

	/* Invoke kernel */
	for(i=0; i<MAX_ITERATIONS; i++) {
		heat_red <<<blocksPerGrid, threadsPerBlock>>> (d_mtx);
		CUT_CHECK_ERROR("Kernel heat_red execution failed");
		heat_black <<<blocksPerGrid, threadsPerBlock>>> (d_mtx);
		CUT_CHECK_ERROR("Kernel heat_black execution failed");
	}
	CUDA_SAFE_CALL(cudaThreadSynchronize());

	/* Stop timer */
	gettimeofday(&end,NULL);
	timeval_subtract(&elapsed, &end, &start);
	secs = (double)elapsed.tv_sec + ((double)elapsed.tv_usec*MICROSEC);

	/* Copy result from device memory to host memory */
	CUDA_SAFE_CALL(cudaMemcpy(h_mtx, d_mtx, size, cudaMemcpyDeviceToHost));

	/* Check for correctness, diagonal neighbors have some heat */
	assert(0.0<h_mtx[OFFBYONE_CORNER]);
	/* After */
	writePPMbinaryImage("image_after.ppm", h_mtx);

	/* Print stats */
	printf("Secs: %f\n", secs);
	printf("GBps: %f\n", ((float)N*N*sizeof(float)*KERNEL_MOP*MAX_ITERATIONS / secs)*1e-9);
	printf("GFlops: %f\n", ((float)N*N*KERNEL_FLOP*MAX_ITERATIONS / secs)*1e-9);

	/* Free device/host memory */
	CUDA_SAFE_CALL(cudaFree(d_mtx)); d_mtx = NULL;
	free(h_mtx); h_mtx = NULL;

	return 0;
}
