#include <stdlib.h> /* calloc */
#include <stdio.h>  /* printf */
#include <assert.h>
#include <sys/time.h>
#include <string.h> /* memcpy */
#include <cuda.h>
#include <cutil_inline.h> /* CUDA_SAFE_CALL */
#include "../../common/common.h"

#define KERNEL_FLOP 4
#define KERNEL_MOP 5

/* Device code */
__global__ void heat(float *oldgrid, float *newgrid)
{
        __shared__ float local[TILE][TILE];
        int i = blockIdx.y * (TILE-2) + threadIdx.y - 1;
        int j = blockIdx.x * (TILE-2) + threadIdx.x - 1;
        int k = threadIdx.y;
	int l = threadIdx.x;

	/* Read */
	if (0<=i && i<N && 0<=j && j<N) {
		local[k][l] = oldgrid[INDEX(i,j)];
	}

	__syncthreads();

	/* Compute */
	if (0<i && i<N-1 && 0<j && j<N-1) {
		if (0<k && k<TILE-1 && 0<l && l<TILE-1) {
			newgrid[INDEX(i,j)] = 0.25f *
				(local[k-1][l] + local[k][l-1] + local[k][l+1] + local[k+1][l]);
		}
	}
}


/* Device code */
__global__ void heat2(float *oldgrid, float *newgrid)
{
 	__shared__ float lold[TILE][TILE];
 	__shared__ float lnew[TILE][TILE];
	int i = blockIdx.y * (TILE-2) + threadIdx.y - 1;	
	int j = blockIdx.x * (TILE-2) + threadIdx.x - 1;
	int k = threadIdx.y;
	int l = threadIdx.x;

	/* Read */
	if (0<=i && i<N && 0<=j && j<N) {
		lold[k][l] = oldgrid[INDEX(i,j)];
	}

	__syncthreads();

	/* Compute */
	if (0<k && k<TILE-1 && 0<l && l<TILE-1) {
		lnew[k][l] = 0.25f *
			(lold[k-1][l] + lold[k][l-1] + lold[k][l+1] + lold[k+1][l]);
	}

	__syncthreads();

	/* Write */
	if (0<i && i<N-1 && 0<j && j<N-1) {
		if (0<k && k<TILE-1 && 0<l && l<TILE-1) {
			newgrid[INDEX(i,j)] = lnew[k][l];
		}
	}
}


/* Host code */
int main(int argc, char** argv)
{
	unsigned int i = 0;
	dim3 threadsPerBlock(TILE,TILE,1);
	dim3 blocksPerGrid(DIV_CEIL(N,TILE-2),DIV_CEIL(N,TILE-2),1);

	/* At least a border square surrounding a single cell */
	assert(2<=N);
	/* New & Old mtx fit into VRAM */
	assert(N*N*sizeof(float) <= (1<<30));

	/* Enough threads */
	assert(N<=blocksPerGrid.x*threadsPerBlock.x);
	assert(N<=blocksPerGrid.y*threadsPerBlock.y);
	/* Hardware limits */
	assert(threadsPerBlock.x*threadsPerBlock.y<=512);
	assert(blocksPerGrid.x<=65535 && blocksPerGrid.y<=65535);

	/* Set device to use */
	CUDA_SAFE_CALL(cudaSetDevice(DEVICE));

	/* Allocate vectors in device memory */
	size_t size = N*N * sizeof(float);
	assert(0<size);

	float *h_old = NULL;
	float *h_new = NULL;
	float *d_old = NULL;
	float *d_new = NULL;

	/* Alloc and set host and device memory */
	h_old = (float *)calloc(1, size);
	assert(h_old);
	h_new = (float *)calloc(1, size);
	assert(h_new);
	CUDA_SAFE_CALL(cudaMalloc((void **) &d_old, size));
	assert(d_old);
	CUDA_SAFE_CALL(cudaMalloc((void **) &d_new, size));
	assert(d_new);
	for (i=0; i<N; i++) {
		float value = HOT - (4.0/(N*N) * HOT)*(N/2.0-i)*(N/2.0-i);
		value = HOT;
		h_old[INDEX(0,i)] = h_old[INDEX(N-1,i)] = value;
		h_old[INDEX(i,0)] = h_old[INDEX(i,N-1)] = value;
	}
	memcpy(h_new, h_old, size);
	/* There is no temp one pixel off the center */
	assert(0.0==h_old[OFFBYONE_CORNER]);
	/* Before */
	writePPMbinaryImage("image_before.ppm", h_old);

	/* Copy vector from host memory to device memory */
	CUDA_SAFE_CALL(cudaMemcpy(d_old, h_old, size, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(d_new, h_new, size, cudaMemcpyHostToDevice));

	/* Start timer */
	double secs = 0.0;
	struct timeval start = {0L,0L}, end = {0L,0L}, elapsed = {0L,0L};
	gettimeofday(&start,NULL);

	/* Invoke kernel */
	for(i=0; i<MAX_ITERATIONS; i++) {
		float *swap = NULL;
		heat <<<blocksPerGrid, threadsPerBlock>>> (d_old, d_new);
		CUT_CHECK_ERROR("Kernel heat execution failed");
		swap = d_old; d_old = d_new; d_new = swap;
	}
	CUDA_SAFE_CALL(cudaThreadSynchronize());

	/* Stop timer */
	gettimeofday(&end,NULL);
	timeval_subtract(&elapsed, &end, &start);
	secs = (double)elapsed.tv_sec + ((double)elapsed.tv_usec*MICROSEC);

	/* Copy result from device memory to host memory */
	CUDA_SAFE_CALL(cudaMemcpy(h_new, d_new, size, cudaMemcpyDeviceToHost));

	/* Check for correctness, diagonal neighbors have some heat */
	assert(0.0<h_new[OFFBYONE_CORNER]);

	/* After */
	writePPMbinaryImage("image_after.ppm", h_new);

	/* Print stats */
	printf("Secs: %f\n", secs);
	printf("GBps: %f\n", ((float)N*N*sizeof(float)*KERNEL_MOP*MAX_ITERATIONS / secs)*1e-9);
	printf("GFlops: %f\n", ((float)N*N*KERNEL_FLOP*MAX_ITERATIONS / secs)*1e-9);

	/* Free device/host memory */
	CUDA_SAFE_CALL(cudaFree(d_old)); d_old = NULL;
	CUDA_SAFE_CALL(cudaFree(d_new)); d_new = NULL;
	free(h_old); h_old = NULL;
	free(h_new); h_new = NULL;

	return 0;
}
