#include <stdlib.h> /* calloc */
#include <stdio.h>  /* printf */
#include <assert.h>
#include <sys/time.h>
#include <string.h> /* memcpy, memset */
#include "../../common/common.h"


/* Within bounds or die */
#define inBounds(i,j) assert(0<=(i) && (i)<N/2 && 0<=(j) && (j)<N)

#define KERNEL_FLOP 4
#define KERNEL_MOP 5

/* Kernel code */
void heat_red(float *red, float *black)
{
	int i = 0, j = 0;
	assert(red && black);
	for (i=0; i<N/2; i++) {
		for (j=1; j<N-1; j++) {
			if (i!=(N-1)/2*(j%2)) { /* not a border in i */
				inBounds(i,j); inBounds(i,j-1); inBounds(i,j+1);
				inBounds(i+(2*(j%2)-1),j);
				red[INDEX(i,j)] = 0.25f * (
					black[INDEX(i,j-1)] + black[INDEX(i,j)] + black[INDEX(i,j+1)] +
					black[INDEX(i+(2*(j%2)-1),j)] /* j%2=0->i-1, j%2=1->i+1 */
							 );
			}
		}
	}
}

void heat_black(float *red, float *black)
{
	int i = 0, j = 0;
	assert(red && black);
	for (i=0; i<N/2; i++) {
		for (j=1; j<N-1; j++) {
			if (i!=(N-1)/2*((j+1)%2)) { /* not a border in i */
				inBounds(i,j); inBounds(i,j-1); inBounds(i,j+1);
				inBounds(i+(2*((j+1)%2)-1),j);
				black[INDEX(i,j)] = 0.25f * (
					red[INDEX(i,j-1)] + red[INDEX(i,j)] + red[INDEX(i,j+1)] +
					red[INDEX(i+(2*((j+1)%2)-1),j)] /* j%2=0->i+1, j%2=1->i-1 */
							   );
			}
		}
	}
}


/* Main code */
int main(int argc, char** argv)
{
	unsigned int i = 0, j = 0;

	/* At least a border square surrounding a single cell */
	assert(2<=N);
	/* Even number of rows for checkboard splitting */
	assert(N%2==0);

	/* Matrix as a vector */
	size_t size = N*N * sizeof(float);
	assert(0<size);	
	float *h_mtx = NULL, *h_tmp = NULL;

	/* Alloc memory */
	h_mtx = (float *)calloc(1, size);
	assert(h_mtx);
	h_tmp = (float *)calloc(1, size);
	assert(h_tmp);

	/* Surroundings on fire */
	for (i=0; i<N; i++) {
		float value = HOT - (4.0/(N*N) * HOT)*(N/2.0-i)*(N/2.0-i);
		h_mtx[INDEX(0,i)] = h_mtx[INDEX(N-1,i)] = value;
		h_mtx[INDEX(i,0)] = h_mtx[INDEX(i,N-1)] = value;
	}

	/* There is no temp one pixel off the center */
	assert(0.0==h_mtx[OFFBYONE_CORNER]);
	/* Before */
	writePPMbinaryImage("image_before.ppm", h_mtx);

	/* Move redblack to red then black */
	for (i=0; i<N; i++) {
		for(j=0; j<N; j++) {
			int in = (((i+j)%2) * N + i)/2;
			assert(0<=in && in<N);
			h_tmp[INDEX(in,j)] = h_mtx[INDEX(i,j)];
		}
	}
	writePPMbinaryImage("image_before_1.ppm", h_tmp);

	/* Start timer */
	double secs = 0.0;
	struct timeval start = {0L,0L}, end = {0L,0L}, elapsed = {0L,0L};
	gettimeofday(&start,NULL);

	/* Invoke kernel */
	for (i=0; i<MAX_ITERATIONS; i++) {
		heat_red(h_tmp, &h_tmp[N*N/2]);
		heat_black(h_tmp, &h_tmp[N*N/2]);
	}

	/* Stop timer */
	gettimeofday(&end,NULL);
	timeval_subtract(&elapsed, &end, &start);
	secs = (double)elapsed.tv_sec + ((double)elapsed.tv_usec*MICROSEC);
	writePPMbinaryImage("image_after_1.ppm", h_tmp);

	/* Move red then black to redblack */
	for (i=0; i<N; i++) {
		for(j=0; j<N; j++) {
			int in = 2*i - N*(i>=N/2) + ((i>=N/2)+j)%2;
			assert(0<=in && in<N);
			h_mtx[INDEX(in,j)] = h_tmp[INDEX(i,j)];
		}
	}

	/* Check for correctness, diagonal neighbors have some heat */
	assert(0.0<h_mtx[OFFBYONE_CORNER]);
	/* After */
	writePPMbinaryImage("image_after.ppm", h_mtx);

	/* Print stats */
	printf("Secs: %f\n", secs);
	printf("GBps: %f\n", ((float)N*N*sizeof(float)*KERNEL_MOP*MAX_ITERATIONS / secs)*1e-9);
	printf("GFlops: %f\n", ((float)N*N*KERNEL_FLOP*MAX_ITERATIONS / secs)*1e-9);

	/* Free memory */
	free(h_mtx); h_mtx = NULL;
	free(h_tmp); h_tmp = NULL;

	return 0;
}
