#include <stdlib.h> /* calloc */
#include <stdio.h>  /* printf, fflush */
#include <assert.h>
#include <sys/time.h>
#include <string.h> /* memcpy, memset */
#include <math.h> /* fabs */
#include <cuda.h>
#include <cutil_inline.h> /* CUDA_SAFE_CALL */
#include "../../common/common.h"

#define KERNEL_FLOP 6
#define KERNEL_MOP 6

/* Device code */
__global__ void heat_red(float *red, float *black, bool *all)
{
	int i = blockIdx.y*TILE + threadIdx.y;
	int j = blockIdx.x*TILE + threadIdx.x;

	float cold = 0.0f, cnew = 0.0f;
	if (0<=i && i<N/2 && 0<j && j<N-1 && i!=(N-1)/2*(j%2)) {
		cold = red[INDEX(i,j)];
		cnew = 0.25f * (
			black[INDEX(i,j-1)] + black[INDEX(i,j)] + black[INDEX(i,j+1)] +
			black[INDEX(i+(2*(j%2)-1),j)]
				);
		red[INDEX(i,j)] = cnew;
	}

	if (EPSILON < fabs(cnew-cold)) {
		*all = false;
	}
}

__global__ void heat_black(float *red, float *black, bool *all)
{
	int i = blockIdx.y*TILE + threadIdx.y;
	int j = blockIdx.x*TILE + threadIdx.x;

	float cold = 0.0f, cnew = 0.0f;
	if (0<=i && i<N/2 && 0<j && j<N-1 && i!=(N-1)/2*((j+1)%2)) {
		cold = black[INDEX(i,j)];
		cnew = 0.25f * (
			red[INDEX(i,j-1)] + red[INDEX(i,j)] + red[INDEX(i,j+1)] +
			red[INDEX(i+(2*((j+1)%2)-1),j)]
				);
		black[INDEX(i,j)] = cnew;
	}
	if (EPSILON < fabs(cnew-cold)) {
		*all = false;
	}
}



/* Host code */
int main(int argc, char** argv)
{
	unsigned int i = 0, j = 0;
	dim3 threadsPerBlock(TILE,TILE,1);
	dim3 blocksPerGrid(DIV_CEIL(N,TILE),DIV_CEIL(N/2,TILE),1);

	/* At least a border square surrounding a single cell */
	assert(2<=N);
	/* Mtx fit into VRAM */
	assert(N*N*sizeof(float) <= (1<<30));
	/* Even number of rows for checkboard splitting */
	assert(N%2==0);

	/* Enough threads */
	assert(N<=blocksPerGrid.x*threadsPerBlock.x);
	assert(N/2<=blocksPerGrid.y*threadsPerBlock.y);
	/* Hardware limits */
	assert(threadsPerBlock.x*threadsPerBlock.y<=512);
	assert(blocksPerGrid.x<=65535 && blocksPerGrid.y<=65535);

	/* Set device to use */
	CUDA_SAFE_CALL(cudaSetDevice(DEVICE));

	/* Allocate vectors in device memory */
	size_t size = N*N * sizeof(float);
	assert(0<size);

	float *h_mtx = NULL;
	float *h_tmp = NULL;
	float *d_red = NULL;
	float *d_black = NULL;
	bool *h_all = NULL;
	bool *d_all = NULL;

	/* Alloc and set host and device memory */
	h_mtx = (float *)calloc(1, size);
	assert(h_mtx);
	h_tmp = (float *)calloc(1, size);
	assert(h_tmp);
	CUDA_SAFE_CALL(cudaMalloc((void **)&d_red, size/2));
	assert(d_red);
	CUDA_SAFE_CALL(cudaMalloc((void **)&d_black, size/2));
	assert(d_black);
	h_all = (bool *)calloc(1, sizeof(bool));
	assert(h_all);
	CUDA_SAFE_CALL(cudaMalloc((void **)&d_all, sizeof(bool)));
	assert(d_all);

	for (i=0; i<N; i++) {
		float value = HOT - (4.0/(N*N) * HOT)*(N/2.0-i)*(N/2.0-i);
		h_mtx[INDEX(0,i)] = h_mtx[INDEX(N-1,i)] = value;
		h_mtx[INDEX(i,0)] = h_mtx[INDEX(i,N-1)] = value;
	}

	/* There is no temp one pixel off the center */
	assert(0.0==h_mtx[OFFBYONE_CORNER]);
	/* Before */
	writePPMbinaryImage("image_before.ppm", h_mtx);

	/* Move redblack to red then black */
	for (i=0; i<N; i++) {
		for(j=0; j<N; j++) {
			int in = (((i+j)%2) * N + i)/2;
			assert(0<=in && in<N);
			h_tmp[INDEX(in,j)] = h_mtx[INDEX(i,j)];
		}
	}
	writePPMbinaryImage("image_before_1.ppm", h_tmp);

	/* Copy vector from host memory to device memory */
	CUDA_SAFE_CALL(cudaMemcpy(d_red, h_tmp, size/2, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(d_black, &h_tmp[N*N/2], size/2, cudaMemcpyHostToDevice));

	/* Start timer */
	double secs = 0.0;
	struct timeval start = {0L,0L}, end = {0L,0L}, elapsed = {0L,0L};
	gettimeofday(&start,NULL);

	/* Invoke kernel */
	unsigned int iters = 0;
	*h_all= false;
	for(iters=0; iters<MAX_ITERATIONS && !*h_all; iters++) {
		*h_all = true;
		CUDA_SAFE_CALL(cudaMemcpy(d_all, h_all, sizeof(bool), cudaMemcpyHostToDevice));
		heat_red <<<blocksPerGrid, threadsPerBlock>>> (d_red, d_black, d_all);
		CUT_CHECK_ERROR("Kernel heat_red execution failed");
		heat_black <<<blocksPerGrid, threadsPerBlock>>> (d_red, d_black, d_all);
		CUT_CHECK_ERROR("Kernel heat_black execution failed");
		CUDA_SAFE_CALL(cudaMemcpy(h_all, d_all, sizeof(bool), cudaMemcpyDeviceToHost));
	}
	CUDA_SAFE_CALL(cudaThreadSynchronize());

	/* Stop timer */
	gettimeofday(&end,NULL);
	timeval_subtract(&elapsed, &end, &start);
	secs = (double)elapsed.tv_sec + ((double)elapsed.tv_usec*MICROSEC);

	/* Copy result from device memory to host memory */
	CUDA_SAFE_CALL(cudaMemcpy(h_tmp, d_red, size/2, cudaMemcpyDeviceToHost));
	CUDA_SAFE_CALL(cudaMemcpy(&h_tmp[N*N/2], d_black, size/2, cudaMemcpyDeviceToHost));

	writePPMbinaryImage("image_after_1.ppm", h_tmp);

	/* After */
	/* Move red then black to redblack */
	for (i=0; i<N; i++) {
		for(j=0; j<N; j++) {
			int in = 2*i - N*(i>=N/2) + ((i>=N/2)+j)%2;
			assert(0<=in && in<N);
			h_mtx[INDEX(in,j)] = h_tmp[INDEX(i,j)];
		}
	}

	/* Check for correctness, diagonal neighbors have some heat */
	assert(0.0<h_mtx[OFFBYONE_CORNER]);

	writePPMbinaryImage("image_after.ppm", h_mtx);

	/* Print stats */
	printf("Iters: %d\n", iters);
	printf("Secs: %f\n", secs);
	printf("GBps: %f\n", ((float)N*N*sizeof(float)*KERNEL_MOP*iters / secs)*1e-9);
	printf("GFlops: %f\n", ((float)N*N*KERNEL_FLOP*iters / secs)*1e-9);

	/* Free device/host memory */
	CUDA_SAFE_CALL(cudaFree(d_red)); d_red = NULL;
	CUDA_SAFE_CALL(cudaFree(d_black)); d_black = NULL;
	CUDA_SAFE_CALL(cudaFree(d_all)); d_all = NULL;
	free(h_mtx); h_mtx = NULL;
	free(h_tmp); h_tmp = NULL;
	free(h_all); h_all = NULL;

	return 0;
}
