#include <stdlib.h> /* calloc */
#include <stdio.h>  /* printf, fflush */
#include <assert.h>
#include <sys/time.h>
#include <string.h> /* memcpy, memset */
#include <math.h> /* fabs */
#include "../../common/common.h"


/* Be in or die */
#define inBounds(i,j) assert(0<=(i) && (i)<N && 0<=(j) && (j)<N)

#define KERNEL_FLOP 7
#define KERNEL_MOP 6

void heat_red(float *mtx, float *diff)
{
	unsigned int i = 0, j = 0;
	assert(mtx && diff);
	for (i=1; i<N-1; i++) {
		for(j=1+((i+1)%2); j<N-1; j+=2) {
			assert((i+j)%2==0); /* is red */
			inBounds(i-1,j); inBounds(i,j-1);
			inBounds(i,j+1); inBounds(i+1,j);
			float cold = mtx[INDEX(i,j)];
			float cnew = 0.25f* (
					mtx[INDEX(i-1,j)] +
				mtx[INDEX(i,j-1)] +	mtx[INDEX(i,j+1)] +
					mtx[INDEX(i+1,j)]
						);
			mtx[INDEX(i,j)] = cnew;
			*diff = MAX(*diff, fabs(cnew - cold));
		}
	}
	return;
}

void heat_black(float *mtx, float *diff)
{
	unsigned int i = 0, j = 0;
	assert(mtx && diff);
	for (i=1; i<N-1; i++) {
		for(j=1+(i%2); j<N-1; j+=2) {
			assert((i+j)%2==1); /* is black */
			inBounds(i-1,j); inBounds(i,j-1);
			inBounds(i,j+1); inBounds(i+1,j);
			float cold = mtx[INDEX(i,j)];
			float cnew = 0.25f* (
					mtx[INDEX(i-1,j)] +
				mtx[INDEX(i,j-1)] +	mtx[INDEX(i,j+1)] +
					mtx[INDEX(i+1,j)]
						);
			mtx[INDEX(i,j)] = cnew;
			*diff = MAX(*diff, fabs(cnew - cold));
		}
	}
	return;
}


/* Main code */
int main(int argc, char** argv)
{
	unsigned int i = 0;

	/* At least a border square surrounding a single cell */
	assert(2<=N);
	/* Even number of rows for checkboard splitting */
	assert(N%2==0);

	/* Matrix as a vector */
	size_t size = N*N * sizeof(float);
	assert(0<size);
	float *h_mtx = NULL;

	/* Alloc memory */
	h_mtx = (float *)calloc(1, size);
	assert(h_mtx);

	/* Surroundings on fire */
	for (i=0; i<N; i++) {
		float value = HOT - (4.0/(N*N) * HOT)*(N/2.0-i)*(N/2.0-i);
		h_mtx[INDEX(0,i)] = h_mtx[INDEX(N-1,i)] = value;
		h_mtx[INDEX(i,0)] = h_mtx[INDEX(i,N-1)] = value;
	}

	/* There is no temp one pixel off the center */
	assert(0.0==h_mtx[OFFBYONE_CORNER]);
	/* Before */
	writePPMbinaryImage("image_before.ppm", h_mtx);

	/* Start timer */
	double secs = 0.0;
	struct timeval start = {0L,0L}, end = {0L,0L}, elapsed = {0L,0L};
	gettimeofday(&start,NULL);

	/* Invoke kernel */
	float diff = EPSILON; 
	unsigned int iters = 0;
	for (iters=0; iters<MAX_ITERATIONS && EPSILON<=diff ; iters++) {
		diff = 0.0f;
		heat_red(h_mtx, &diff);
		heat_black(h_mtx, &diff);
	}

	/* Stop timer */
	gettimeofday(&end,NULL);
	timeval_subtract(&elapsed, &end, &start);
	secs = (double)elapsed.tv_sec + ((double)elapsed.tv_usec*MICROSEC);

	/* Check for correctness, diagonal neighbors have some heat */
	assert(0.0<h_mtx[OFFBYONE_CORNER]);

	writePPMbinaryImage("image_after.ppm", h_mtx);

	/* Print stats */
	printf("Iters: %d\n", iters);
	printf("Secs: %f\n", secs);
	printf("GBps: %f\n", ((double)N*N*sizeof(float)*KERNEL_MOP*iters / secs)*1e-9);
	printf("GFlops: %f\n", ((double)N*N*KERNEL_FLOP*iters / secs)*1e-9);

	/* Free memory */
	free(h_mtx); h_mtx = NULL;

	return 0;
}
