#include <stdlib.h> /* malloc */
#include <stdio.h>  /* printf */
#include <assert.h>
#include <sys/time.h>
#include <cuda.h>
#include <cutil_inline.h> /* CUDA_SAFE_CALL */
#include "../../common/common.h"


/* Device code */
__global__ void saxpy(unsigned long int M, float a, float *x, float *y)
{
#define KERNEL_FLOP 2
#define KERNEL_MOP 3
	int i = blockIdx.x*blockDim.x+threadIdx.x;
	if (i < M) {
		y[i] += a*x[i];
	}
}


/* Host code */
int main(int argc, char** argv)
{
	unsigned int M = 0;
	unsigned int threadsPerBlock = 0;
	unsigned int blocksPerGrid = 0;

	// Handle input params
	if (argc<3 || argc>4) {
		printf("Usage: %s 2^<bits> <threadsPerBlock> [<blocksPerGrid>]\n", argv[0]);
		exit(1);
	}
	assert(argc==3 || argc==4);
	int i;
	for(i=0; i<argc; i++) {
		assert(argv[i]);
	}
	M = 1 << atoi(argv[1]);
	threadsPerBlock = atoi(argv[2]);
	if (argc==4) {
		blocksPerGrid = atoi(argv[3]);
	} else {
		blocksPerGrid = DIV_CEIL(M,threadsPerBlock);
	}

	/* Hardware limits */
	assert(threadsPerBlock<=512);
	assert(blocksPerGrid<=65535);
	assert(blocksPerGrid*threadsPerBlock>=M);

	/* Allocate vectors in device memory */
	size_t size = M * sizeof(float);
	assert(0<size);

	float a = 2.0;
	float *h_x = NULL, *h_y = NULL;
	float *d_x = NULL, *d_y = NULL;

	/* Alloc and set host and device memory */
	h_x = (float *)malloc(size);
	assert(h_x);
	h_y= (float *)malloc(size);
	assert(h_y);
	CUDA_SAFE_CALL(cudaMalloc((void **) &d_x, size));
	assert(d_x);
	CUDA_SAFE_CALL(cudaMalloc((void **) &d_y, size));
	assert(d_y);
	for (i=0; i<M; i++) {
		h_x[i] = h_y[i] = 2.0;
	}

	/* Copy vectors from host memory to device memory
	 * h_x and h_y are input vectors stored in host memory
	 */
	CUDA_SAFE_CALL(cudaMemcpy(d_x, h_x, size, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(d_y, h_y, size, cudaMemcpyHostToDevice));

	/* Start timer */
	double secs = 0.0;
	struct timeval start = {0L,0L}, end = {0L,0L}, elapsed = {0L,0L};
	gettimeofday(&start, NULL);

	/* Invoke kernel */
	saxpy <<<blocksPerGrid, threadsPerBlock>>> (M, a, d_x, d_y);
	CUT_CHECK_ERROR("Kernel saxpy execution failed");
	CUDA_SAFE_CALL(cudaThreadSynchronize());

	/* Stop timer */
	gettimeofday(&end, NULL);
	timeval_subtract(&elapsed, &end, &start);
	secs = (double)elapsed.tv_sec + ((double)elapsed.tv_usec*MICROSEC);

	/* Copy result from device memory to host memory
	 * h_C contains the result in host memory
	 */
	cudaMemcpy(h_y, d_y, size, cudaMemcpyDeviceToHost);

	/* Check for correctness */
	for (i=0; i<M; i++) {
		assert(h_y[i]==2*h_x[i]+h_x[i]);
	}

	/* Print results */
	printf("Secs: %f\n", secs);
	printf("GBps: %f\n", ((float)M*sizeof(float)*KERNEL_MOP / secs)*1e-9);
	printf("Flops: %f\n", ((float)M*KERNEL_FLOP / secs)*1e-9);

	/* Free device/host memory */
	cudaFree(d_x); d_x = NULL;
	cudaFree(d_y); d_y = NULL;
	free(h_x); h_x = NULL;
	free(h_y); h_y = NULL;

	return 0;
}
